# scanner that takes folder structure into account

from ..db import db,Audio,Artwork,Album,Artist,Track

from doreah.io import NestedProgressBar, ProgressBar

import os
from .. import cleanup
import yaml
#from doreah.settings import get_settings
#from doreah.io import NestedProgressBar



def build_metadata(filelist,trees):

	prog_build = ProgressBar(
		len(filelist),
		prefix="Parsing metadata    "
	)

	todo = []
	if len(filelist) > 0:
		for t in trees:
			td, _ = scan_tree_for(filelist,t,prog_build)
			todo += td



	for af in todo:
		if af.track.artists == []:
			af.track.artists = [Artist("Various Artists")]
		if af.track.albums == []:
			Album(albumartists=af.track.artists,name=af.track.title).tracks.append(af.track)
		prog_build.progress(step=af.track.title)

	prog_build.done()

	for alb in db.getall(Album):
		if alb.albumartists == []:
			artists = {}
			for t in alb.tracks:
				for a in t.artists:
					artists[a] = artists.setdefault(a,0) + 1
			artists_in_album = [a for a in artists]
	#		artists_in_album.sort(key=lambda x:artists[x],reverse=True)
			commonartists = [a for a in artists_in_album if artists[a] > len(alb.tracks)/2]
			if len(commonartists) == 0:
				commonartists = [Artist(name="Various Artists")]
			albnew = Album(albumartists=commonartists,name=alb.name)
			albnew.artworks = alb.artworks

			for t in alb.tracks:
			 	albnew.tracks.append(t)
			db.delete(alb)



	prog_imag = ProgressBar(
		len(filelist),
		prefix="Extracting artwork  "
	)


	for aud in filelist:
		prog_imag.progress()
		for aw in aud.get_embedded_artworks()["album"]:
			if aw not in aud.track.albums[0].artworks:
					aud.track.albums[0].artworks.append(aw)



	newly_created_tracks = [f.track for f in filelist]
	affected_albums = list(set([t.albums[0] for t in newly_created_tracks]))

	prog_alb = ProgressBar(
		len(affected_albums),
		prefix="Fixing album orders "
	)

	for alb in affected_albums:
		tracks = list(set(alb.tracks))
		numstracks = [(t.audiofiles[0].metadata()["position"],t) for t in tracks]
		numstracks.sort(key=lambda x:x[0])
		alb.tracks = [nt[1] for nt in numstracks]
		prog_alb.progress()


def scan_tree_for(files,tree,progressbar):
	subdirs,audiofiles,imagefiles = tree.subdirs, tree.audiofiles, tree.imagefiles

	# sort into ones we can use to gather info about the folder
	# and the ones we want to find out metadata about
	todo_audiofiles, scanned_audiofiles = [],[]
	for sd in subdirs:
		td,sc = scan_tree_for(files,sd,progressbar)
		todo_audiofiles += td
		scanned_audiofiles += sc

	#print("todo",[af.path for af in todo_audiofiles],"scanned",[af.path for af in scanned_audiofiles])


	for af in audiofiles:
		if af in files: todo_audiofiles.append(af)
		elif af.track: scanned_audiofiles.append(af)
		else: print(af,"has no track, but will not be scanned!")



	possible_folder_artists = {}
	possible_folder_albums = {}
	possible_folder_albumartists = {}


	# check if we even need to find out anything about this folder
	if tree.artist is None or tree.album is None:

		for af in scanned_audiofiles:
			track = af.track
			artists = track.artists
			for artist in artists:
				possible_folder_artists[artist] = possible_folder_artists.setdefault(artist,0) + 1
				possible_folder_albumartists[artist] = possible_folder_albumartists.setdefault(artist,0) # set to 0 so it's there
			try:
				album = track.albums[0]
				albumartists = album.albumartists

				possible_folder_albums[album] = possible_folder_albums.setdefault(album,0) + 1
				for artist in albumartists:
					possible_folder_albumartists[artist] = possible_folder_albumartists.setdefault(artist,0) + 1
					possible_folder_artists[artist] = possible_folder_artists.setdefault(artist,0) # set to 0 so it's there
			except:
				#print(track,"has no albums")
				pass







	# scan the files
	for af in todo_audiofiles[:]:
		metadata = af.metadata()
		artists,title = cleanup.fullclean(metadata["artists"],metadata["title"])
		albumartists,album = cleanup.cleanartists([metadata["albumartist"]]),metadata["album"]
		pos,dur = metadata["position"],metadata["length"]


		tr = Track(
			title=title,
			artists=[Artist(name=a) for a in artists],
			length=dur
		)
		tr.audiofiles.append(af)



		if tree.artist is None or tree.album is None:
			for artist in tr.artists:
				possible_folder_artists[artist] = possible_folder_artists.setdefault(artist,0) + 1
				possible_folder_albumartists[artist] = possible_folder_albumartists.setdefault(artist,0)

		if album not in [None,""]:

			al = Album(
				name=album,
				albumartists=[Artist(name=a) for a in albumartists]
			)
			al.tracks.append(tr)

			if tree.artist is None or tree.album is None:
				for artist in al.albumartists:
					possible_folder_albumartists[artist] = possible_folder_albumartists.setdefault(artist,0) + 1
					possible_folder_artists[artist] = possible_folder_artists.setdefault(artist,0)
				possible_folder_albums[al] = possible_folder_albums.setdefault(al,0) + 1

			if albumartists != [] and artists != []:
				# metadata fully parsed from embedded data, we're good!
				progressbar.progress(step=af.path)
				scanned_audiofiles.append(af)
				todo_audiofiles.remove(af)
				# this also means this will below be counted to find out folder album / artist
				continue



	total_files = len(todo_audiofiles) + len(scanned_audiofiles)

	# if not, now try to determine folder album / artist
	if tree.artist is None:
		artistlist = list(set(a for a in possible_folder_artists))
		artistlist.sort(key=lambda x:possible_folder_artists[x]+possible_folder_albumartists[x],reverse=True)
		if len(artistlist) > 0 and possible_folder_artists[artistlist[0]]+possible_folder_albumartists[artistlist[0]] >= total_files/1.5:
			tree.artist = artistlist[0]

	if tree.album is None:
		albumlist = [a for a in possible_folder_albums]
		albumlist.sort(key=lambda x:possible_folder_albums[x],reverse=True)
		if len(albumlist) > 0 and possible_folder_albums[albumlist[0]] >= total_files/1.5:
			#print("common album",commonalbum)
			tree.album = albumlist[0]
			#print("assigned",albumlist[0],"to folder, artist is",tree.artist)
			#if tree.album.albumartists in [[],"",None]:
			#	if tree.artist is not None:
			#		#tree.album.albumartists = [tree.artist]
			#		al = Album(name=tree.album.name,albumartists=[tree.artist])
			#		print("new album",al)
			#		al.tracks += tree.album.tracks
			#		tree.album.tracks = []
			#		tree.album = al
					# set folder artist as albumartist




	# assign folder album / artist to tracks that are still missing it
	for af in todo_audiofiles[:]:
		if af.track.artists == [] and tree.artist is not None:
			af.track.artists = [tree.artist]
		if af.track.albums == [] and tree.album is not None:
			tree.album.tracks.append(af.track)


		if af.track.albums != [] and af.track.albums[0].albumartists == [] and \
				tree.album is not None and af.track.albums[0].name == tree.album.name:
			af.track.albums[0].tracks.remove(af.track)
			tree.album.tracks.append(af.track)


	#	if af.track.albums != [] and af.track.albums[0].albumartists == [] and tree.artist is not None:
	#		al = Album(name=af.track.albums[0].name,albumartists=[tree.artist])
	#		af.track.albums[0].tracks.remove(af.track)
	#		al.tracks.append(af.track)


		if af.track.artists != [] and af.track.albums != [] and af.track.albums[0].albumartists != []:
			progressbar.progress(step=af.path)
			scanned_audiofiles.append(af)
			todo_audiofiles.remove(af)



	## check artwork files
	for i in imagefiles:
		if "artist" in i.path.lower() and tree.artist is not None and i not in tree.artist.artworks:
			tree.artist.artworks.append(i)
		elif "album" in i.path.lower() and tree.album is not None and i not in tree.album.artworks:
 			tree.album.artworks.append(i)


	return todo_audiofiles, scanned_audiofiles



import os
from .__init__ import DATA_DIR
os.chdir(DATA_DIR)

from bottle import Bottle, run
import waitress
from . import web
from . import db
from .apis import subsonic
from threading import *
import time
from doreah.settings import get_settings
from doreah import auth
#import auth
import signal


HOST, PORT = get_settings("HOST","PORT")
THREADS = 12





#db.load_database()
db.scan_new()



def graceful_exit(sig=None,frame=None):
	print("Server shutting down...")
	db.save_database()
	os._exit(42)

signal.signal(signal.SIGINT, graceful_exit)
signal.signal(signal.SIGTERM, graceful_exit)


server = Bottle()

db.api.mount(server=server)
auth.authapi.mount(server=server)
subsonic.api.mount(server=server)
web.server_handlers(server=server)

#run(server, host='::', port=PORT, server='waitress',threads=24)
waitress.serve(server,host=HOST,port=PORT,threads=THREADS)



### PACKAGE DATA

__name__ = "albula"
__desc__ = "A minimalist self-hosted music server."
__version__ = 0,5,2
__versionstr__ = ".".join(str(n) for n in __version__)
__author__ = {
	"name":"Johannes Krattenmacher",
	"email":"albula@krateng.dev",
	"github":"krateng"
}

__requires__ = [
	"bottle>=0.12.16",
	"waitress>=1.3",
	"doreah>=1.2.10",
	"nimrodel>=0.6.1",
	"mutagen>=1.42",
	"pyyaml>=5.1",
	"lesscpy>=0.13"
]
__resources__ = [
	"web/*/*",
	"web/*",
	"static/*/*",
	"data_files/*/*"
]

__commands__ = {
	"albula":"controller:main"
}

### DOREAH CONFIGURATION

from doreah import config
config(
	logging={
		"logfolder": "logs"
	},
	settings={
		"files":[
			"settings/default.ini",
			"settings/settings.ini"
		]
	},
	auth={
		"multiuser": False,
		"cookieprefix": "albula"
	}
)


### USER DATA FOLDER


import os
import pkg_resources
from distutils import dir_util

try:
	DATA_DIR = os.environ["XDG_DATA_HOME"].split(":")[0]
	assert os.path.exists(DATA_DIR)
except:
	DATA_DIR = os.path.join(os.environ["HOME"],".local/share/")

DATA_DIR = os.path.join(DATA_DIR,"albula")
os.makedirs(DATA_DIR,exist_ok=True)



def copy_initial_local_files():
	folder = pkg_resources.resource_filename(__name__,"data_files")
	#shutil.copy(folder,DATA_DIR)
	dir_util.copy_tree(folder,DATA_DIR)

copy_initial_local_files()



from doreah.settings import get_settings
from doreah.io import NestedProgressBar, ProgressBar
import os
import yaml
from ..db import Audio, Artwork, Artist, Album, Track, db


AUDIOFORMATS = ["mp3","flac"]
IMAGEFORMATS = ["jpeg","jpg","png","webp"]

class Directory:
	def __init__(self,name,subdirs=[],files=[]):
		self.name = name
		self.subdirs = list(subdirs)
		self.audiofiles = []
		self.imagefiles = []
		for f in files:
			if isinstance(f,Audio): self.audiofiles.append(f)
			elif isinstance(f,Artwork): self.imagefiles.append(f)

		self.artists = {}
		self.albums = {}
		self.artist = None
		self.album = None

	def add_file(self,f):
		if isinstance(f,Audio): self.audiofiles.append(f)
		elif isinstance(f,Artwork): self.imagefiles.append(f)

	def total_files(self):
		num = len(self.imagefiles) + len(self.audiofiles)
		for sd in self.subdirs:
			num += sd.total_files()
		return num

# build directory tree
def scan_dir(dir,progressbar):
	d = Directory(name=dir.split("/")[-1])

	dirlist = os.listdir(dir)
	progressbar.godeeper(len(dirlist))
	for f in dirlist:
		fullpath = os.path.join(dir,f)

		if f in get_settings("IGNORE_FILES"):
			break
		if f.startswith("."):
			progressbar.progress(step=f)
			continue
		if os.path.isdir(fullpath):
			d.subdirs.append(scan_dir(fullpath,progressbar))
		else:
			ext = f.split(".")[-1].lower()
			if ext in AUDIOFORMATS: d.add_file(Audio(path=fullpath))
			elif ext in IMAGEFORMATS: d.add_file(Artwork(path=fullpath))
			elif f in ["artist.yml","artist.yaml"]:
				with open(fullpath,"r") as fil:
					#files.append({**yaml.safe_load(fil),"type":"artist"})
					info = yaml.safe_load(fil)
					d.artist = Artist(name=info["name"])
			elif f in ["album.yml","album.yaml"]:
				with open(fullpath,"r") as fil:
					#files.append({**yaml.safe_load(fil),"type":"album"})
					info = yaml.safe_load(fil)
					d.album = Album(name=info["name"],albumartists=[Artist(name=a) for a in info["albumartists"]])

			#else: print("File",f,"has unknown format")
			progressbar.progress(step=f)

	progressbar.done()
	return d


def scan(dirs):
	old_audiofiles = db.getall(Audio)
	for a in old_audiofiles:
		if not os.path.exists(a.path):
			a.track.audiofiles.remove(a)
			db.delete(a)


	scan_dir_prog = NestedProgressBar(len(dirs),prefix="Scanning directories",manual=True)
	trees = []
	new_audiofiles = []
	for dir in dirs:
		d = scan_dir(dir,progressbar=scan_dir_prog)
		trees.append(d)

	new_audiofiles = [a for a in db.getall(Audio) if a not in old_audiofiles]

	old_tracks = db.getall(Track)

	#when scanning, also call metadata refresher for newly added files
	from ..dbbuild.parser import build_metadata
	build_metadata(new_audiofiles,trees)

	new_tracks = [t for t in db.getall(Track) if t not in old_tracks]

	# import maloja data for new tracks
	malserver = get_settings("MALOJA_SERVER")
	if malserver is not None:
		from ..dbbuild.malojareader import get_scrobbles
		get_scrobbles(new_tracks)


	from ..dbbuild.pruner import prune_database
	prune_database()

	return trees



# non-relational db

from nimrodel import EAPI
import requests
import os
import re

import random


import mutagen
from mutagen.mp3 import MP3
from mutagen.flac import FLAC
import zlib


#from db_oo_helper import Ref, MultiRef, DBObject, db, save_database, load_database
from doreah.database import Database, Ref, MultiRef
from doreah.settings import get_settings
from doreah.io import ProgressBar

db = Database(file="database.ddb")


class Artwork(db.DBObject):
	__primary__ = "path",
	path: str

	def read(self):
		if self.path.lower().endswith(".jpeg") or self.path.lower().endswith(".jpg"):
			mime = 'image/jpeg'
		elif self.path.lower().endswith(".png"):
			mime = 'image/png'
		elif self.path.lower().endswith(".webp"):
			mime = 'image/webp'
		with open(self.path,"rb") as imagefile:
			stream = imagefile.read()

		return mime,stream

	def link(self):
		return "/artwork/" + str(self.uid)

# if no artwork is present, return default file
class PA:
	def __init__(self,type):
		self.type = type
		self.uid = -1
	def link(self):
		return "/artwork/" + self.type

pseudoartworks = {"artist":PA("artist"),"album":PA("album"),"track":PA("track")}
def PseudoArtwork(type):
	return pseudoartworks[type]




class Audio(db.DBObject):
	__primary__ = "path",
	path: str

	def read(self):
		if self.path.endswith(".mp3"):
			mime = 'audio/mp3'
		elif self.path.endswith(".flac"):
			mime = 'audio/flac'
		with open(self.path,"rb") as audiofile:
			stream = audiofile.read()

		return mime,stream


	# extracts artwork, saves it and returns reference
	def get_embedded_artworks(self):
		ext = self.path.split(".")[-1].lower()
		if ext in ["flac"]:
			audio = FLAC(self.path)
			imgs = audio.pictures

		elif ext in ["mp3"]:
			audio = MP3(self.path)
			tags = audio.tags
			imgs = tags.getall("APIC")

		artworks = {"album":[],"track":[],"artist":[]}
		for i in imgs:

			data = i.data
			imghash = str(zlib.adler32(data))
			mime = i.mime

			imagefile = imghash[:3] + "/" + imghash[3:] + "." + mime.split("/")[-1]
			artwork = Artwork(path="cache/" + imagefile)
			if not os.path.exists("cache/" + imagefile):
				os.makedirs("cache/" + imagefile.split("/")[0],exist_ok=True)
				with open("cache/" + imagefile,"wb") as fi:
					fi.write(data)


			if i.type == 3: artworks["album"].append(artwork)

		return artworks

	def metadata(self):
		try:
			return self.cached_metadata
		except:
			pass

		ext = self.path.split(".")[-1].lower()
		if ext in ["flac"]:
			audio = FLAC(self.path)

			tags = audio.tags
			try:
				album = [entry[1] for entry in tags if entry[0] == "ALBUM"][0]
			except:
				album = None
			try:
				title = [entry[1] for entry in tags if entry[0] == "TITLE"][0]
			except:
				title = None
			artists = [entry[1] for entry in tags if entry[0] == "ARTIST"]
			try:
				albumartist = [entry[1] for entry in tags if entry[0] == "ALBUMARTIST"][0]
			except:
				albumartist = None
			try:
				pos = [entry[1] for entry in tags if entry[0] == "TRACKNUMBER"][0]
			except:
				pos = 0

			length = int(audio.info.length)

		elif ext in ["mp3"]:
			audio = MP3(self.path)

			tags = audio.tags
			try:
				album = tags.get("TALB").text[0]
			except:
				album = None
			try:
				title = tags.get("TIT2").text[0]
			except:
				title = None

			#artists = [set(obj.text) for obj in tags.getall("TPE1") + tags.getall("TPE2") + tags.getall("TPE3") + tags.getall("TPE4")]
			artists = [set(obj.text) for obj in tags.getall("TPE1")]
			artists = set.union(*artists)
			try:
				albumartist = tags.get("TPE2").text[0]
			except:
				albumartist = None

			try:
				pos = tags.get("TRCK").text[0].split("/")[0]
			except:
				pos = 0

			length = int(audio.info.length)



		if title is None or pos is 0:
			filename = self.path.split("/")[-1].split(".")[0]
			num, name = re.match(r"([0-9]*)[ -]*(.*)",filename).groups()
			if pos == 0 and num != "": pos = int(num)
			if title is None: title = name

		self.cached_metadata = {
			"title":title,
			"artists":artists,
			"album":album,
			"albumartist":albumartist,
			"length":length,
			"position":int(pos)

		}

		return self.cached_metadata


class Artist(db.DBObject):
	__primary__ = "name",
	__dbsettings__ = {"ignore_capitalization":True}
	name: str
	artworks: list = MultiRef(Artwork,exclusive=False,backref="artist")
	artwork_index: int

	def __apidict__(self):
		return {
			"uid":self.uid,
			"name":self.name,
			#"sorttitle":self.name.lower(),
		#	"artwork":[a.uid for a in self.artwork],
			"artwork":self.get_artwork().link(),
			"artwork_choices":[a.link() for a in self.artworks],
			"last_played":max([t.lastplayed for t in self.tracks] + [0]),
			"times_played":sum(t.timesplayed for t in self.tracks),
			"length":sum(t.length for t in self.tracks),
			"track_ids":[track.uid for track in self.tracks]
		}

	def get_artwork(self):
		if len(self.artworks) > 0: return self.artworks[self.artwork_index]
		else: return PseudoArtwork("artist")

	def get_tracklist(self):
		return self.tracks





class Track(db.DBObject):
	__primary__ = "title","artists"
	__dbsettings__ = {"ignore_capitalization":True}
	title: str
	artists: list = MultiRef(Artist,backref="tracks",exclusive=False)
	#albums: list = MultiRef(Album,backref="tracks",exclusive=False)
	audiofiles: list = MultiRef(Audio,exclusive=True,backref="track")
	artworks: list = MultiRef(Artwork,exclusive=False,backref="track")
	artwork_index: int
	length: int
	lastplayed: int
	timesplayed: int

	def __apidict__(self):
		return {
			"uid":self.uid,
			"title":self.title,
			"length":self.length,
			#"sorttitle":self.title.lower(),
			"artists":[{"id":a.uid,"name":a.name} for a in self.artists],
			"albums":[{"id":a.uid,"name":a.name} for a in self.albums],
		#	"artist_ids":list(a.uid for a in self.artists),
		#	"artist_names":list(a.name for a in self.artists),
		#	"artwork":[a.uid for a in self.artwork],
			"artwork":self.get_artwork().link(),
			"artwork_choices":[a.link() for a in self.artworks],
			"last_played":self.lastplayed,
			"times_played":self.timesplayed,
			"track_ids":[self.uid]
		}

	def get_artwork(self):
		if len(self.artworks) > 0: return self.artworks[self.artwork_index]
		else:
			for a in self.albums:
				if len(a.artworks) > 0: return a.artworks[a.artwork_index]
		return PseudoArtwork("track")

	def get_audio(self):
		if len(self.audiofiles) > 0: return self.audiofiles[0]
		return None

	def get_tracklist(self):
		return [self]



class Album(db.DBObject):
	__primary__ = "name","albumartists"
	__dbsettings__ = {"ignore_capitalization":True}
	name: str
	albumartists: list = MultiRef(Artist,exclusive=False,backref="albums")
	tracks: list = MultiRef(Track,exclusive=False,backref="albums")
	artworks: list = MultiRef(Artwork,exclusive=False,backref="album")
	artwork_index: int

	def __apidict__(self):
		return {
			"uid":self.uid,
			"albumartists":[{"id":a.uid,"name":a.name} for a in self.albumartists],
		#	"albumartist_ids":list(a.uid for a in self.albumartists),
		#	"albumartist_names":list(a.name for a in self.albumartists),
			"name":self.name,
			#"sorttitle":self.name.lower(),
		#	"artwork":[a.uid for a in self.artwork],
			"artwork":self.get_artwork().link(),
			"artwork_choices":[a.link() for a in self.artworks],
			"last_played":max([t.lastplayed for t in self.tracks] + [0]),
			"times_played":sum(t.timesplayed for t in self.tracks),
			"length":sum(t.length for t in self.tracks),
			"track_ids":[track.uid for track in self.tracks]
		}

	def get_artwork(self):
		if len(self.artworks) > 0: return self.artworks[self.artwork_index]
		else: return PseudoArtwork("album")



from doreah import auth

api = EAPI(path="api",delay=True,auth=auth.check)


@api.get("artists")
def list_artists():
	return db.getall(Artist)

@api.get("albums")
def list_albums():
	return db.getall(Album)

@api.get("tracks")
def list_tracks():
	return db.getall(Track)


@api.get("artist/{id}")
def get_artist(id:int):
	artist = db.get(id)
	result = {}
	result["artist"] = artist
	result["tracks"] = list(artist.tracks)
	result["albums"] = list(artist.albums)
	for t in result["tracks"]:
		result["albums"] += t.albums
	result["albums"] = list(set(result["albums"]))

	return result


@api.get("album/{id}")
def get_album(id:int):
	album = db.get(id)
	result = {}
	result["album"] = album
	result["tracks"] = list(album.tracks)
	# sort artists by presence on this album
	artisttracks = {}
	for t in result["tracks"]:
		for a in t.artists:
			artisttracks[a] = artisttracks.setdefault(a,0) + 1
	result["artists"] = sorted([a for a in artisttracks],key=lambda x: artisttracks[x],reverse=True)
	#result["artists"] = list(set(a for t in result["tracks"] for a in t.artists))

	return result

@api.get("track/{id}")
def get_track(id:int):
	track = db.get(id)
	result = {}
	result["track"] = track
	result["artists"] = list(track.artists)
	result["albums"] = list(track.albums)

	return result


def get_file_by_ref(uid):
	#print("getting file path for uid",uid)
	path = db.get(uid).path
	return path

def get_artwork_of(uid):
	obj = db.get(uid)
	artwork = obj.get_artwork()
	return artwork


@api.post("play")
def play_track(id:int,seconds:int,time:int):
	track = db.get(id)
	if seconds > 5:
		track.timesplayed += 1
		track.lastplayed = time

	if seconds > (track.length / 2):

		if get_settings("MALOJA_SCROBBLE"):
			server = get_settings("MALOJA_SERVER")
			key = get_settings("MALOJA_KEY")
			url = server + "/api/newscrobble"
			data = {
				"artist":"/".join(a.name for a in track.artists),
				"title":track.title,
				"duration":seconds,
				"time":time,
				"key":key
			}

		requests.post(url, data=data)

@api.post("setartwork")
def set_artwork(element:int,artwork:int):
	element = db.get(element)
	artwork = db.get(artwork)
	assert artwork in element.artworks
	element.artwork_index = element.artworks.index(artwork)

@api.post("setname")
def set_name(element:int,name:str):
	element = db.get(element)
	if isinstance(element,Artist) or isinstance(element,Album):
		element.name = name
	elif isinstance(element,Track):
		element.title = name




from .dbbuild.scanner import scan
from .dbbuild.parser import build_metadata
from .dbbuild.pruner import prune_database

@api.post("scan")
def scan_new():
	dirs = get_settings("MUSIC_DIRECTORIES")
	trees = scan(dirs)


@api.post("refresh_metadata")
def refresh_metadata(audio_ids=None):
	if audio_ids is None:
		audios = db.getall(Audio)
	else:
		audios = [db.get(id) for id in audio_ids]

	dirs = get_settings("MUSIC_DIRECTORIES")
	trees = scan(dirs)
	build_metadata(audios,trees)


@api.post("clean_database")
def clean_database():
	prune_database()


	# remove file references that have no objects
#	artworks = db.getall(Artwork)
#	referenced_artworks = set().union(*[set(entity.artworks) for entity in db.getall(Track) + db.getall(Album) + db.getall(Artist)])
#	for a in artworks:
#		if a not in referenced_artworks:
#			print(a.path,"no longer referenced, removing...")
#			db.delete(a)
#
#	audiofiles = db.getall(Audio)
#	referenced_audiofiles = set().union(*[set(entity.audiofiles) for entity in db.getall(Track)])
#	for a in audiofiles:
#		if a not in referenced_audiofiles:
#			print(a.path,"no longer referenced, removing...")
#			db.delete(a)
#
#	# remove tracks that have no audiofiles
#	tracks = db.getall(Track)
#	for t in tracks:
#		if t.audiofiles == []:
#			print(t,"has no music file associated, removing...")
#			db.delete(t)

def save_database():
	db.save()



from bottle import Bottle, route, get, post, error, run, template, static_file, request, response, FormsDict, redirect, template
from importlib.machinery import SourceFileLoader
from doreah.pyhp import file as pyhpfile
from . import db
#import auth
from doreah import auth
import pkg_resources
import os


WEBFOLDER = pkg_resources.resource_filename(__name__,"web")
STATICFOLDER = pkg_resources.resource_filename(__name__,"static")

pthjoin = os.path.join

def generate_css():
	import lesscpy
	from io import StringIO
	less = ""
	for f in os.listdir(pthjoin(STATICFOLDER,"less")):
		with open(pthjoin(STATICFOLDER,"less",f),"r") as lessf:
			less += lessf.read()

	css = lesscpy.compile(StringIO(less),minify=True)
	return css

css = generate_css()

def server_handlers(server):


	@server.get("/style.css")
	def get_css():
		response.content_type = 'text/css'
		return css

	@server.get("/<name>.<ext>")
	def file(name,ext):
		return static_file(ext + "/" + name + "." + ext,root=STATICFOLDER)


	@server.get("/")
	def start():
		if auth.check(request):
			result = pyhpfile(os.path.join(WEBFOLDER,"main.pyhp"),{"db":db})
		else:
			#result = pyhpfile("web/login.pyhp",{"db":db,"auth":auth})
			result = auth.get_login_page(stylesheets=["/style.css"])

		return result


#	@server.get("/imgof/<uid>")
#	@auth.authenticated
#	def imageof(uid):
#		uid = int(uid)
#
#		obj = db.db.get(uid)
#		artwork = obj.get_artwork()
#		if artwork is None: return ""
#
#		mime,stream = artwork.read()
#		response.set_header('Content-type', mime)
#		response.set_header("Cache-Control", "public, max-age=360000")
#		return stream

	@server.get("/artwork/<uid>")
	@auth.authenticated
	def image(uid):
		try:
			uid = int(uid)
		except:
			return static_file("png/unknown_" + uid + ".png",root=STATICFOLDER)

		artwork = db.db.get(uid)
		mime,stream = artwork.read()
		response.set_header('Content-type', mime)
		response.set_header("Cache-Control", "public, max-age=360000")
		return stream


	@server.get("/audioof/<uid>")
	@auth.authenticated
	def audioof(uid):

		uid = int(uid)

		obj = db.db.get(uid)
		audio = obj.get_audio()
		if audio is None: return ""

		mime,stream = audio.read()
		response.set_header('Content-type', mime)
		response.set_header("Cache-Control", "public, max-age=360000")
		return stream



from doreah.settings import get_settings
from doreah.io import print_progress_bar
import requests
import urllib




def get_scrobbles(tracks):

	tracks = {(frozenset(a.name.lower() for a in t.artists),t.title.lower()):t for t in tracks}
	matched_tracks = {}

	malserver = get_settings("MALOJA_SERVER")
	if malserver is None:
		malserver = input("Please enter the URL of your Maloja server: ")

	# list of tracks
	url = malserver + "/api/tracks"
	result = requests.get(url)
	result = result.json()["list"]
	for r in result:
		tr = (frozenset(a.lower() for a in r["artists"]),r["title"].lower())
		if tr in tracks:
			matched_tracks[tracks[tr]] = r

	#print(matched_tracks)


	url = malserver + "/api/trackinfo"
	i = 0
	tracknumber = len(matched_tracks)
	for t in matched_tracks:
		i += 1
		trackinfo = matched_tracks[t]
		info = []
		for a in trackinfo["artists"]: info.append(("artist",a))
		info.append(("title",trackinfo["title"]))
	#	trackurl = url + "?" + "&".join(["artist=" + a for a in trackinfo["artists"]]) + "&title=" + trackinfo["title"]
		trackurl = url + "?" + urllib.parse.urlencode(info)
	#	print("Making request",trackurl)
		result = requests.get(trackurl)
		result = result.json()
		#print(t,"has",result["scrobbles"],"scrobbles")
		print_progress_bar(num=(i,tracknumber),prefix="Importing scrobbles ")
		t.timesplayed = result["scrobbles"]

	



from nimrodel import EAPI
from doreah import auth, settings
import hashlib
from .. import db



def subsonicauth(request):

	try:
		token = request.query.get("t")
		salt = request.query.get("s")

		key = settings.get_settings("SUBSONIC_KEY")

		m = hashlib.md5()
		m.update(key.encode())
		m.update(salt.encode())

		if m.hexdigest() == token:
#			print("Success!")
			return True

		assert False
	except:
		raise
		return False



api = EAPI(path="subsonicapi/rest",delay=True,auth=subsonicauth,type="xml",root_node="subsonic-response",debug=True)

base_response = {
	"xmlns":"http://subsonic.org/restapi",
	"status":"ok",
	"version":"1.13.0"
}

@api.get("ping.view")
def ping(**kwargs):
	return base_response

@api.get("getUser.view")
def getuser(**kwargs):
	return {**base_response, **{
		"user":{
			"username":"The User",
			"email":"Who knows",
			"scrobblingEnabled":"true",
			"adminRole":"false",
			"settingsRole":"false",
			"downloadRole":"true",
			"uploadRole":"false",
			"playlistRole":"false",
			"coverArtRole":"false",
			"commentRole":"false",
			"podcastRole":"false",
			"streamRole":"false",
			"jukeboxRole":"false",
			"shareRole":"false"
		}
	}}

@api.get("getRandomSongs.view")
def getrandomsongs(size:int=10,**kwargs):
	return {**base_response, **{
		"randomSongs":{
			"song":[
				{"id":t.uid,"title":t.title,"album":t.albums[0].name,"artist":", ".join(a.name for a in t.artists)}
				for t in db.db.getall(db.Track)[:size]
			]
		}
	}}

@api.get("getArtists.view")
def getartists(**kwargs):
	return {**base_response, **{
		"artists":{
			"index":[
				{"name":char,
				"artist":[
					{"id":a.uid,"name":a.name,"coverArt":a.get_artwork().uid,"albumCount":len(a.albums)}
					for a in db.db.getall(db.Artist) if a.name.startswith(char)
				]}
				for char in "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
			]
		}
	}}

@api.get("getCoverArt.view")
def getcoverart(id:int,**kwargs):
	stream = db.db.get(id).get_artwork().read()[1]
	print(stream)



import os
from .__init__ import DATA_DIR
os.chdir(DATA_DIR)

import getpass
from doreah.control import mainfunction


def setpassword():
	print("Input new password")
	pw = getpass.getpass()
	print("Repeat please")
	if pw != getpass.getpass():
		print("Passwords do not match!")
	else:
		from doreah import auth
		auth.defaultuser.setpw(pw)

def updateplays():
	from .db import db, Track
	from .dbbuild.malojareader import get_scrobbles


	tracks = db.getall(Track)
	get_scrobbles(tracks)
	db.save()

@mainfunction({},shield=True)
def main(action,**kwargs):
	actions = {
		"setpassword":setpassword,
		"updateplays":updateplays
	}

	return actions[action](**kwargs)



from ..db import db, Track, Album, Artist, Audio, Artwork
from doreah.io import ProgressBar

def prune_database():

	alltracks = db.getall(Track)
	allalbums = db.getall(Album)
	allartists = db.getall(Artist)

	allartworks = db.getall(Artwork)
	allaudios = db.getall(Audio)

	prog = ProgressBar(
		len(alltracks) + len(allalbums) + len(allartists) + len(allartworks) + len(allaudios),
		prefix="Pruning database    "
	)

	referenced_audiofiles = set()
	referenced_artworks = set()

	for t in alltracks:
		# remove track if it has no audio files
		if t.audiofiles == []:
			#print(t,"has no music file associated, removing...")
			db.delete(t)

		# list all refrenced files
		for af in t.audiofiles:
			referenced_audiofiles.add(af)
		for aw in t.artworks:
			referenced_artworks.add(aw)

		prog.progress()

	for e in allalbums + allartists:
		# remove album / artist if no tracks / albums
		if e.tracks == [] and getattr(e,"albums",[]) == []:
			#print(e,"has no tracks, removing...")
			db.delete(e)

		for aw in e.artworks:
			referenced_artworks.add(aw)

		prog.progress()


	for a in allartworks:
		if a not in referenced_artworks:
			#print(a.path,"no longer referenced, removing...")
			db.delete(a)
		prog.progress()
	for a in allaudios:
		if a not in referenced_audiofiles:
			#print(a.path,"no longer referenced, removing...")
			db.delete(a)
		prog.progress()

	prog.done()



import re
import unicodedata
from doreah.settings import get_settings
import yaml
import os

def fullclean(artists_,title):
	artists = []
	for a in artists_:
		artists += parseArtists(removespecial(a))
	#artists = parseArtists(removespecial(artist))
	title = parseTitle(removespecial(title))
	(title,moreartists) = parseTitleForArtists(title)
	artists += moreartists
	artists = list(set(artists))
	artists.sort()

	return (artists,title)

def cleanartists(artists_):
	artists = []
	for a in artists_:
		artists += parseArtists(removespecial(a))
	artists = list(set(artists))
	artists.sort()
	return artists

def removespecial(s):
	if s is None: return ""
	s = s.replace("\t","").replace("␟","").replace("\n","")
	s = re.sub(" +"," ",s)
	return "".join(ch for ch in s if unicodedata.category(ch)[0]!="C")



delimiters_feat = ["ft.","ft","feat.","feat","featuring","Ft.","Ft","Feat.","Feat","Featuring"]			#Delimiters used for extra artists, even when in the title field
delimiters = ["vs.","vs","&"]											#Delimiters in informal artist strings, spaces expected around them.
delimiters = get_settings("ARTIST_DELIMITERS")							#risk of false positives. user has control over tagging
delimiters_formal = ["; ",";","/"]										#Delimiters used specifically to tag multiple artists when only one tag field is available, no spaces used

safe_artists = []
for f in os.listdir("rules"):
	if not f.startswith("."):
		with open("rules/" + f,"r") as fil:
			d = yaml.safe_load(fil)
			safe_artists += d["artists"]

def parseArtists(a):

	if a.strip() == "":
		return []

#		if a.strip() in rules_notanartist:
#			return []

	if " performing " in a.lower():
		return parseArtists(re.split(" [Pp]erforming",a)[0])

	if a.strip() in safe_artists:
		return [a.strip()]
#		if a.strip() in rules_replaceartist:
#			return rules_replaceartist[a.strip()].split("␟")



	for d in delimiters_feat:
		if re.match(r"(.*) \(" + d + " (.*)\)",a) is not None:
			return parseArtists(re.sub(r"(.*) \(" + d + " (.*)\)",r"\1",a)) + parseArtists(re.sub(r"(.*) \(" + d + " (.*)\)",r"\2",a))

	for d in delimiters_formal:
		if (d in a):
			ls = []
			for i in a.split(d):
				ls += parseArtists(i)
			return ls

	for d in (delimiters_feat + delimiters):
		if ((" " + d + " ") in a):
			ls = []
			for i in a.split(" " + d + " "):
				ls += parseArtists(i)
			return ls





	return [a.strip()]

def parseTitle(t):
#		if t.strip() in rules_replacetitle:
#			return rules_replacetitle[t.strip()]

	t = t.replace("[","(").replace("]",")")

	t = re.sub(r" \(as made famous by .*?\)","",t)
	t = re.sub(r" \(originally by .*?\)","",t)
	t = re.sub(r" \(.*?Remaster.*?\)","",t)

	return t.strip()

def parseTitleForArtists(t):
	for d in delimiters_feat:
		if re.match(r"(.*) \(" + d + " (.*?)\)",t) is not None:
			(title,artists) = parseTitleForArtists(re.sub(r"(.*) \(" + d + " (.*?)\)",r"\1",t))
			artists += parseArtists(re.sub(r"(.*) \(" + d + " (.*?)\).*",r"\2",t))
			return (title,artists)
		if re.match(r"(.*) - " + d + " (.*)",t) is not None:
			(title,artists) = parseTitleForArtists(re.sub(r"(.*) - " + d + " (.*)",r"\1",t))
			artists += parseArtists(re.sub(r"(.*) - " + d + " (.*).*",r"\2",t))
			return (title,artists)
		if re.match(r"(.*) " + d + " (.*)",t) is not None:
			(title,artists) = parseTitleForArtists(re.sub(r"(.*) " + d + " (.*)",r"\1",t))
			artists += parseArtists(re.sub(r"(.*) " + d + " (.*).*",r"\2",t))
			return (title,artists)

	return (t,[])
